/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              reLOC 0.09-vegas                              */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// multirobot.h / 0.09-vegas_092
/*----------------------------------------------------------------------------*/
//
// Multirobot coordinated path-finding solving package.
//
/*----------------------------------------------------------------------------*/


#ifndef __MULTIROBOT_H__
#define __MULTIROBOT_H__

#include <vector>
#include <list>
#include <set>
#include <map>

#include "types.h"
#include "result.h"
#include "reloc.h"
#include "cnf.h"
#include "graph.h"


using namespace std;
using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/
// sRobotArrangement

    class sRobotArrangement
    {
    public:
	static const int VACANT_VERTEX = 0; /* Robots are numbered starting with 1 */
	static const int UNDEFINED_LOCATION = -1;

    public:
	typedef std::vector<int> Robots_vector;
	typedef std::vector<int> Vertices_vector;

    public:
	sRobotArrangement();
	sRobotArrangement(int N_Vertices, int N_Robots, bool random = false);
	sRobotArrangement(const sRobotArrangement &robot_arrangement);
	const sRobotArrangement& operator=(const sRobotArrangement &robot_arrangement);

	bool operator==(const sRobotArrangement &robot_arrangement) const;
	bool operator<(const sRobotArrangement &robot_arrangement) const;

	int get_RobotCount(void) const;
	int get_VertexCount(void) const;

	int get_RobotLocation(int robot_id) const;
	int get_VertexOccupancy(int vertex_id) const;

	void place_Robot(int robot_id, int vertex_id);
	void remove_Robot(int robot_id);
	void clean_Vertex(int vertex_id);

	void move_Robot(int robot_id, int dest_vertex_id);
	bool verify_Move(int robot_id, int dest_vertex_id, const sUndirectedGraph &graph) const;

	void randomize(void);

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Stream(FILE *fw, const sString &indent = "") const;

	virtual sResult to_File_multirobot(const sString &filename, const sString &indent = "") const;
	virtual void to_Stream_multirobot(FILE *fw, const sString &indent = "") const;

	virtual sResult from_File_multirobot(const sString &filename, int component = 0);
	virtual sResult from_Stream_multirobot(FILE *fr, int component = 0);

    public:
	Robots_vector m_robot_Locs;
	Vertices_vector m_vertex_Occups;
    };


/*----------------------------------------------------------------------------*/
// sRobotGoal

    class sRobotGoal
    {
    public:
	static const int UNDEFINED_LOCATION = sRobotArrangement::UNDEFINED_LOCATION;

    public:
	typedef std::set<int> Vertices_set;
	typedef std::vector<Vertices_set> Goals_vector;

	typedef std::set<int> Robots_set;
	typedef std::vector<Robots_set> Compats_vector;

    public:
	sRobotGoal();
	sRobotGoal(int N_Vertices, int N_Robots);
	sRobotGoal(int N_Vertices, int N_Robots, int N_Goals);
	sRobotGoal(const sRobotArrangement &robot_arrangement);
	sRobotGoal(const sRobotGoal &robot_goal);
	const sRobotGoal& operator=(const sRobotGoal &robot_goal);

	int get_RobotCount(void) const;
	int get_VertexCount(void) const;

	const Vertices_set& get_RobotGoal(int robot_id) const;
	Vertices_set& provide_RobotGoal(int robot_id);

	const Robots_set& get_GoalCompatibility(int goal_id) const;
	Robots_set& provide_GoalCompatibility(int goal_id);

	void charge_Robot(int robot_id, int goal_id);
	void charge_Robot(int robot_id, const Vertices_set &goal_IDs);

	void assign_Goal(int goal_id, int robot_id);
	void assign_Goal(int goal_id, const Robots_set &robot_IDs);

	void discharge_Robot(int robot_id);
	void discharge_Robot(int robot_id, int goal_id);
	void discharge_Robot(int robot_id, const Vertices_set &goal_IDs);

	bool is_Satisfied(const sRobotArrangement &robot_arrangement) const;

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Stream(FILE *fw, const sString &indent = "") const;

	virtual sResult to_File_multirobot(const sString &filename, const sString &indent = "") const;
	virtual void to_Stream_multirobot(FILE *fw, const sString &indent = "") const;
	virtual void to_Stream_multirobot(FILE *fw, const Robots_set &robot_IDs, const sString &indent = "") const;

	virtual sResult from_File_multirobot(const sString &filename, int component = 0);
	virtual sResult from_Stream_multirobot(FILE *fr, int component = 0);
	virtual sResult from_Stream_multirobot(FILE *fr, Robots_set &robot_IDs);

    public:
	Goals_vector m_robot_Goals;
	Compats_vector m_goal_Compats;
    };


/*----------------------------------------------------------------------------*/
// sMultirobotCNFEncodingContext
    
    class sMultirobotEncodingContext_CNFsat
    {	
    public:
	static const int UNDEFINED_LAYER_COUNT;

    public:
	enum GeneratingMode
	{
	    GENERATING_STANDARD,
	    GENERATING_ADVANCED,
	    GENERATING_BITWISE
	};

    public:
	typedef std::vector<sIndexableStateIdentifier> StateIdentifiers_vector;
	typedef std::vector<sIndexableBitIdentifier> BitIdentifiers_vector;
	typedef std::vector<StateIdentifiers_vector> StateIdentifiers_2d_vector;

	typedef std::map<int, sIndexableStateIdentifier*, std::less<int> > StateIdentifiers_map;
	typedef std::map<int, sIndexableBitIdentifier*, std::less<int> > BitIdentifiers_map;

    public:
	sMultirobotEncodingContext_CNFsat();
	sMultirobotEncodingContext_CNFsat(int N_Layers);
	sMultirobotEncodingContext_CNFsat(const sMultirobotEncodingContext_CNFsat &encoding_context);
	const sMultirobotEncodingContext_CNFsat& operator=(const sMultirobotEncodingContext_CNFsat &encoding_context);

	GeneratingMode get_GeneratingMode(void) const;
	
	void switch_GeneratingMode(GeneratingMode generating_mode);
	void switchTo_StandardGeneratingMode(void);
	void switchTo_AdvancedGeneratingMode(void);
	void switchTo_BitwiseGeneratingMode(void);

	void register_InternalIdentifiers(void);
	void register_TranslateIdentifier(sIndexableStateIdentifier &state_identifier);
	void register_TranslateIdentifier(sIndexableBitIdentifier &bit_identifier);

	sSpecifiedIdentifier translate_CNF_Variable(int cnf_variable) const;

    public:
	int m_N_Layers;
	sVariableStore_CNF m_variable_store;

	GeneratingMode m_generating_mode;

	sStateClauseGenerator *m_clause_generator;
	sStateClauseGenerator m_state_clause_generator;
	sAdvancedClauseGenerator m_advanced_clause_generator;
	sBitwiseClauseGenerator m_bitwise_clause_generator;

	sBitClauseGenerator *m_bit_generator;
	sBitClauseGenerator m_bit_clause_generator;

	sIndexableStateIdentifier m_vertex_occupancy;
	sIndexableStateIdentifier m_robot_location;
	sIndexableStateIdentifier m_transition_action;
	sIndexableBitIdentifier m_vertex_occupancy_by_water;
	sIndexableBitIdentifier m_vertex_occupancy_by_robot;

	BitIdentifiers_vector m_edge_occupancy_by_water;
	StateIdentifiers_vector m_transition_Actions;

	StateIdentifiers_map m_state_Identifiers;
	BitIdentifiers_map m_bit_Identifiers;
    };


/*----------------------------------------------------------------------------*/
// sMultirobotInstance

    class sMultirobotInstance
    {
    public:
	enum GoalType
	{
	    GOAL_TYPE_UNDEFINED,
	    GOAL_TYPE_ARRANGEMENT,
	    GOAL_TYPE_SPECIFICATION
	};

    public:
	sMultirobotInstance();
	sMultirobotInstance(const sUndirectedGraph &environment, const sRobotArrangement &initial_arrangement, const sRobotArrangement &goal_arrangement);
	sMultirobotInstance(const sUndirectedGraph &environment, const sRobotArrangement &initial_arrangement, const sRobotGoal &goal_specification);
	sMultirobotInstance(const sMultirobotInstance &multirobot_instance);
	const sMultirobotInstance& operator=(const sMultirobotInstance &multirobot_instance);

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Screen_multirobot(const sString &indent = "") const;
	virtual void to_Screen_domainPDDL(const sString &indent = "") const;
	virtual void to_Screen_problemPDDL(const sString &indent = "") const;
	virtual void to_Screen_InverseCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_AdvancedCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_DifferentialCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_BijectionCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_HeuristicDifferentialCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Screen_HeuristicBijectionCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Screen_HeuristicAdvancedCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Screen_PuzzleCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_BitwiseCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Screen_FlowCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_MatchingCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Screen_HeuristicMatchingCNFsat(sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);

	virtual sResult to_File(const sString &filename, const sString &indent = "") const;
	virtual sResult to_File_multirobot(const sString &filename, const sString &indent = "") const;
	virtual sResult to_File_domainPDDL(const sString &filename, const sString &indent = "") const;
	virtual sResult to_File_problemPDDL(const sString &filename, const sString &indent = "") const;
	virtual sResult to_File_InverseCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_AdvancedCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_DifferentialCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_BijectionCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_HeuristicDifferentialCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual sResult to_File_HeuristicBijectionCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual sResult to_File_HeuristicAdvancedCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual sResult to_File_PuzzleCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_BitwiseCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual sResult to_File_FlowCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_MatchingCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual sResult to_File_HeuristicMatchingCNFsat(const sString &filename, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);

	virtual void to_Stream(FILE *fw, const sString &indent = "") const;
	virtual void to_Stream_multirobot(FILE *fw, const sString &indent = "") const;
	virtual void to_Stream_domainPDDL(FILE *fw, const sString &indent = "") const;
	virtual void to_Stream_problemPDDL(FILE *fw, const sString &indent = "") const;
	virtual void to_Stream_InverseCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_AdvancedCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_DifferentialCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_BijectionCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_HeuristicDifferentialCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Stream_HeuristicBijectionCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Stream_HeuristicAdvancedCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Stream_PuzzleCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_BitwiseCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);
	virtual void to_Stream_FlowCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_MatchingCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false) const;
	virtual void to_Stream_HeuristicMatchingCNFsat(FILE *fw, sMultirobotEncodingContext_CNFsat &encoding_context, const sString &indent = "", bool verbose = false);

    public:
	GoalType m_goal_type;

	sUndirectedGraph m_environment;
	sRobotArrangement m_initial_arrangement;
	sRobotArrangement m_goal_arrangement;
	sRobotGoal m_goal_specification;
    };


/*----------------------------------------------------------------------------*/
// sMultirobotSolution

    class sMultirobotSolution
    {
    public:
	static const int N_STEPS_UNDEFINED = -1;

    public:
	struct Move
	{
	    Move(int robot_id, int src_vrtx_id, int dest_vrtx_id);

	    bool is_Undefined(void) const;

	    int m_robot_id;
	    int m_src_vrtx_id;
	    int m_dest_vrtx_id;
	};

	static const Move UNDEFINED_MOVE;

	typedef std::list<Move> Moves_list;

	struct Step
	{
	    Step(int time);

	    int m_time;
	    Moves_list m_Moves;
	};

	typedef std::vector<Step> Steps_vector;
	
    public:
	sMultirobotSolution();
	sMultirobotSolution(int start_step, const sMultirobotSolution &sub_solution);
	sMultirobotSolution(const sMultirobotSolution &sub_solution_1, const sMultirobotSolution sub_solution_2);
	sMultirobotSolution(const sMultirobotSolution &multirobot_solution);
	const sMultirobotSolution& operator=(const sMultirobotSolution &multirobot_solution);

	int get_MoveCount(void) const;
	int get_StepCount(void) const;

	void add_Move(int time, const Move &move);

	int calc_EmptySteps(void) const;
	void remove_EmptySteps(void);

	sMultirobotSolution extract_Subsolution(int start_step, int final_step) const;

	void execute_Solution(const sRobotArrangement &initial_arrangement,
			      sRobotArrangement       &final_arrangement,
			      int                      N_Steps = N_STEPS_UNDEFINED) const;

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Stream(FILE *fw, const sString &indent = "") const;

	virtual sResult to_File_multirobot(const sString &filename, const sString &indent = "") const;
	virtual void to_Stream_multirobot(FILE *fw, const sString &indent = "") const;

	virtual sResult from_File_multirobot(const sString &filename);
	virtual sResult from_Stream_multirobot(FILE *fr);

    public:
	int m_Moves_cnt;
	Steps_vector m_Steps;
    };


/*----------------------------------------------------------------------------*/
// sMultirobotFlowModel

    class sMultirobotFlowModel
    {
    public:
	typedef std::vector<int> Robots_vector;

    public:
	sMultirobotFlowModel(const sMultirobotInstance &instance);

	void build_Network(int N_steps);
	void build_Network(const Robots_vector &robot_selection, int N_steps);
	void destroy_Network(void);

	int compute_Relocation(void);

	int compute_Distance(void);
	int compute_Distance(const Robots_vector &robot_selection);
	int compute_Distance(int *N_tries);

    public:
	sMultirobotInstance m_instance;
	sDigraph m_flow_network;

	sDigraph::Vertices_map::iterator m_network_source;
	sDigraph::Vertices_map::iterator m_network_sink;
    };



/*----------------------------------------------------------------------------*/

} // namespace sReloc

#endif /* __MULTIROBOT_H__ */
